"""user.py contains basic information about the steno3d user"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties


class User(properties.PropertyClass):
    """Class representing a user instance"""
    _model_api_location = "user"
    email = properties.String('email')
    name = properties.String('name')
    url = properties.String('url')
    affiliation = properties.String('affiliation')
    location = properties.String('location')



"""base.py sets up BaseExample class that allows examples to be accessed
as properties at the class level
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import realpath
from os.path import sep

from future.utils import with_metaclass


class exampleproperty(object):
    """wrapper that sets class method as saved property"""

    def __init__(self, func):
        self.func = classmethod(func)
        self.propname = '_' + func.__name__

    def __get__(self, cls, owner):
        if getattr(owner, self.propname, None) is None:
            setattr(owner, self.propname, self.func.__get__(None, owner)())
        return getattr(owner, self.propname)


class exampleoutput(object):
    """wrapper that sets class method to remove saved properties"""

    def __init__(self, func, directory):
        self.func = classmethod(func)
        self._directory = ['_' + prop for prop in directory]
        self._stash = [None for prop in directory]

    def __get__(self, cls, owner):
        def _output():
            """return the original function surrounded by stash and unstash"""
            self.stash(owner)
            out = self.func.__get__(None, owner)()
            self.unstash(owner)
            return out
        return _output

    def stash(self, owner):
        """stash saved properties during example output"""
        for i, prop in enumerate(self._directory):
            self._stash[i] = getattr(owner, prop, None)
            setattr(owner, prop, None)

    def unstash(self, owner):
        """unstash saved properties after example output"""
        for i, prop in enumerate(self._directory):
            setattr(owner, prop, self._stash[i])


class _ExampleMetaClass(type):
    """metaclass that wraps every function with exampleproperty()"""

    def __new__(mcs, name, bases, attrs):
        if name not in ('BaseExample', 'Images'):
            assert ('get_resources' in attrs and
                    callable(attrs['get_resources'])), \
                'Example {cls} class must have method get_resources()'.format(
                    cls=name
                )
            assert ('get_project' in attrs and
                    callable(attrs['get_project'])), \
                'Example class {cls} must have method get_project()'.format(
                    cls=name
                )

        example_directory = []
        for attr in attrs:
            value = attrs[attr]
            if not attr.startswith('get_') and callable(value):
                attrs[attr] = exampleproperty(value)
                example_directory.append(attr)

        for attr in attrs:
            value = attrs[attr]
            if attr.startswith('get_') and callable(value):
                attrs[attr] = exampleoutput(value, example_directory)

        return type.__new__(mcs, name, bases, attrs)


class BaseExample(with_metaclass(_ExampleMetaClass, object)):
    """basic class that all examples inherit from"""

    def __init__(self, *args, **kwargs):
        raise TypeError('Examples cannot be instantiated. Please access '
                        'the properties directly.')

    def asset_dir(self):
        """path to directory containing all assets"""
        return sep.join(realpath(__file__).split(sep)[:-3] + ['assets'])



"""surface.py contains the Surface composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from numpy import max as npmax
from numpy import min as npmin
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh2DOptions(MeshOptions):
    pass


class _SurfaceOptions(ColorOptions):
    pass


class Mesh2D(BaseMesh):
    """class steno3d.Mesh2D

    Contains spatial information about a 2D surface defined by
    triangular faces.
    """
    vertices = properties.Array(
        'Mesh vertices',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    triangles = properties.Array(
        'Mesh triangle vertex indices',
        shape=('*', 3),
        dtype=int,
        required=True
    )
    opts = properties.Pointer(
        'Mesh2D Options',
        ptype=_Mesh2DOptions
    )

    def __init__(self, *args, **kwargs):
        super(Mesh2D, self).__init__(*args, **kwargs)
        self._nN = None
        self._nC = None

    @property
    def nN(self):
        """ get number of nodes """
        if getattr(self, '_nN', None) is None:
            self._nN = len(self.vertices)
        return self._nN

    @property
    def nC(self):
        """ get number of cells """
        if getattr(self, '_nC', None) is None:
            self._nC = len(self.triangles)
        return self._nC

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if npmin(self.triangles) < 0:
            raise ValueError('Triangles may only have positive integers')
        if npmax(self.triangles) >= len(self.vertices):
            raise ValueError('Triangles expects more vertices than provided')
        return True

    @property
    def _files(self):
        """Get the file information"""
        files = {
            'triangles':
                self._properties['triangles'].serialize(self.triangles),
            'vertices': self._properties['vertices'].serialize(self.vertices),
        }
        return files


class Mesh2DGrid(BaseMesh):
    """Contains spatial information of a 2D grid."""
    h1 = properties.Array(
        'Grid cell widths, x-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h2 = properties.Array(
        'Grid cell widths, y-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    x0 = properties.Vector(
        'Origin vector',
        default=[0, 0, 0]
    )
    Z = properties.Array(
        'Node topography',
        shape=('*',),
        dtype=float
    )
    opts = properties.Pointer(
        'Mesh2D Options',
        ptype=_Mesh2DOptions
    )

    def __init__(self, *args, **kwargs):
        super(Mesh2DGrid, self).__init__(*args, **kwargs)
        self._nN = None
        self._nC = None

    @property
    def nN(self):
        """ get number of nodes """
        if getattr(self, '_nN', None) is None:
            self._nN = (len(self.h1)+1) * (len(self.h2)+1)
        return self._nN

    @property
    def nC(self):
        """ get number of cells """
        if getattr(self, '_nC', None) is None:
            self._nC = len(self.h1) * len(self.h2)
        return self._nC

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.x0.nV != 1:
            raise ValueError('Origin x0 must be only one vector')
        if getattr(self, 'Z', None) is not None and len(self.Z) != self.nN:
            raise ValueError(
                'Length of Z, {zlen}, must equal number of nodes, '
                '{nnode}'.format(
                    zlen=len(self.Z),
                    nnode=self.nN
                )
            )
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""
        datadict = {
            'tensors': dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
            )),
            'OUV': dumps(dict(
                O=self.x0.tolist()[0],
                U=[self.h1.sum(), 0, 0],
                V=[0, self.h2.sum(), 0],
            )),
        }
        return datadict

    @property
    def _files(self):
        """Get the file information"""
        if getattr(self, 'Z', None) is None:
            return {}
        files = {
            'Z': self._properties['Z'].serialize(self.Z),
        }
        return files


class _SurfaceBinder(properties.PropertyClass):
    """Contains the data on a 2D surface with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('FACE', 'CELLCENTER'),
            'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Surface(CompositeResource):
    """Contains all the information about a 2D surface"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=[Mesh2D, Mesh2DGrid],
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_SurfaceBinder,
        repeated=True
    )
    textures = properties.Pointer(
        'Textures',
        ptype='Texture2DImage',
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_SurfaceOptions
    )

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'surface.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Surface, self).validate()
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""

        datadict = {
            'mesh': dumps({
                'uid': self.mesh.json_after_upload['longUid'],
            }),
            'data': dumps([
                {
                    'location': d.location,
                    'uid': d.data.json_after_upload['longUid'],
                } for d in self.data
            ]),
            'textures': dumps([
                {
                    'uid': t.json_after_upload['longUid'],
                } for t in self.textures
            ]),
        }
        return datadict


__all__ = ['Surface', 'Mesh2D', 'Mesh2DGrid']



"""wolfpass.py is a project example using resources from Wolf Pass"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import sep

from glob import glob
from numpy import load as npload

from .base import BaseExample
from ..data import DataArray
from ..line import Line
from ..line import Mesh1D
from ..point import Mesh0D
from ..point import Point
from ..project import Project
from ..surface import Mesh2D
from ..surface import Surface
from ..texture import Texture2DImage
from ..volume import Mesh3DGrid
from ..volume import Volume


class Wolfpass(BaseExample):

    def wolfpass_dir(self):
        """base wolfpass directory"""
        return sep.join([self.asset_dir, 'wolfpass'])

    # Borehole Collar Points:
    def collar_point_dir(self):
        """directory for collar points"""
        return sep.join([self.wolfpass_dir, 'collar_point'])

    def collar_vertices(self):
        """collar point vertices"""
        return npload(sep.join([self.collar_point_dir, 'drill_loc_v.npy']))

    def collar_data(self):
        """collar raw data

        list of dicts of 'location' and 'data' with 'title' and 'array'
        """
        raw_data = []
        for npyfile in glob(sep.join([self.collar_point_dir, '*.npy'])):
            if npyfile.endswith('_v.npy'):
                continue
            raw_data += [dict(
                location='N',
                data=DataArray(
                    title=npyfile.split(sep)[-1][:-4],
                    array=npload(npyfile)
                )
            )]
        return raw_data

    def collar_points(self):
        """Steno3D point resource for borehole collars"""
        return Point(
            project=self.project,
            mesh=Mesh0D(
                vertices=self.collar_vertices
            ),
            data=self.collar_data,
            title='Borehole Collar Locations'
        )

    # Borehole Lines:
    def borehole_line_dir(self):
        """directory for borehole lines"""
        return sep.join([self.wolfpass_dir, 'borehole_line'])

    def borehole_vertices(self):
        """borehole line vertices"""
        return npload(sep.join([self.borehole_line_dir, 'boreholes_v.npy']))

    def borehole_segments(self):
        """borehole segment vertex indices"""
        return npload(sep.join([self.borehole_line_dir, 'boreholes_s.npy']))

    def borehole_data(self):
        """borehole raw data

        list of dicts of 'location' and 'data' with 'title' and 'array'
        """
        raw_data = []
        for npyfile in glob(sep.join([self.borehole_line_dir, '*.npy'])):
            if npyfile.endswith('_v.npy') or npyfile.endswith('_s.npy'):
                continue
            raw_data += [dict(
                location='CC',
                data=DataArray(
                    title=npyfile.split(sep)[-1][:-4],
                    array=npload(npyfile)
                )
            )]
        return raw_data

    def borehole_lines(self):
        """Steno3D line resource for boreholes"""
        return Line(
            project=self.project,
            mesh=Mesh1D(
                vertices=self.borehole_vertices,
                segments=self.borehole_segments
            ),
            data=self.borehole_data,
            title='Boreholes'
        )

    # CU Percentage Surfaces:
    def cu_surface_dir(self):
        """directory for CU percent surfaces"""
        return sep.join([self.wolfpass_dir, 'cu_surface'])

    def cu_prefixes(self):
        """list of prefixes for the different cu pct surfaces"""
        return [full_file[:-6] for full_file in
                glob(sep.join([self.cu_surface_dir, '*_v.npy']))]

    def cu_vertices(self):
        """list of cu pct surface vertices"""
        return [npload(prefix + '_v.npy') for prefix in self.cu_prefixes]

    def cu_triangles(self):
        """list of cu pct surface triangles"""
        return [npload(prefix + '_t.npy') for prefix in self.cu_prefixes]

    def cu_surfaces(self):
        """tuple of  Steno3D surface resources for each CU percent range"""
        cu_surfs = tuple()
        for i, prefix in enumerate(self.cu_prefixes):
            cu_surfs += (Surface(
                project=self.project,
                mesh=Mesh2D(
                    vertices=self.cu_vertices[i],
                    triangles=self.cu_triangles[i]
                ),
                title=prefix.split(sep)[-1]
            ),)
        return cu_surfs

    # Lithology Surfaces:
    def lith_surface_dir(self):
        """directory for lithology surfaces"""
        return sep.join([self.wolfpass_dir, 'lith_surface'])

    def lith_prefixes(self):
        """list of prefixes for the different lithology surfaces"""
        return [full_file[:-6] for full_file in
                glob(sep.join([self.lith_surface_dir, '*_v.npy']))]

    def lith_vertices(self):
        """list of lithology surface vertices"""
        return [npload(prefix + '_v.npy') for prefix in self.lith_prefixes]

    def lith_triangles(self):
        """list of lithology surface triangles"""
        return [npload(prefix + '_t.npy') for prefix in self.lith_prefixes]

    def lith_diorite_early_data(self):
        """data for early diorite surface"""
        npyfile = sep.join([self.lith_surface_dir, 'dist_to_borehole.npy'])
        return dict(
            location='N',
            data=DataArray(
                title=npyfile.split(sep)[-1][:-4],
                array=npload(npyfile)
            )
        )

    def lith_surfaces(self):
        """tuple of  Steno3D surface resources for each CU percent range"""
        lith_surfs = tuple()
        for i, prefix in enumerate(self.lith_prefixes):
            if prefix.split(sep)[-1] == 'diorite_early':
                lith_data = [self.lith_diorite_early_data]
            else:
                lith_data = []
            lith_surfs += (Surface(
                project=self.project,
                mesh=Mesh2D(
                    vertices=self.lith_vertices[i],
                    triangles=self.lith_triangles[i]
                ),
                data=lith_data,
                title=prefix.split(sep)[-1]
            ),)
        return lith_surfs

    # Topography Surfaces:
    def topo_surface_dir(self):
        """directory for topography surface"""
        return sep.join([self.wolfpass_dir, 'topo_surface'])

    def topo_vertices(self):
        """topography vertices"""
        return npload(sep.join([self.topo_surface_dir, 'topo_v.npy']))

    def topo_triangles(self):
        """topography triangles"""
        return npload(sep.join([self.topo_surface_dir, 'topo_t.npy']))

    def topo_texture(self):
        """topography surface image texture"""
        return Texture2DImage(
            O=[443200, 491750, 0],
            U=[4425, 0, 0],
            V=[0, 3690, 0],
            image=sep.join([self.topo_surface_dir, 'topography.png'])
        )

    def topo_data(self):
        """elevation data"""
        return dict(
            location='N',
            data=DataArray(
                title='elevation',
                array=npload(sep.join([self.topo_surface_dir,
                                       'Elevation.npy']))
            )
        )

    def topo_surface(self):
        """Steno3D surface resource with topo data and surface imagery"""
        return Surface(
            project=self.project,
            mesh=Mesh2D(
                vertices=self.topo_vertices,
                triangles=self.topo_triangles
            ),
            data=self.topo_data,
            textures=self.topo_texture,
            title='Topography Surface'
        )

    # Cross-section Surface:
    def xsect_surface_dir(self):
        """directory for cross section surfaces"""
        return sep.join([self.wolfpass_dir, 'xsect_surface'])

    def xsect_vertices(self):
        """cross section vertices"""
        return npload(sep.join([self.xsect_surface_dir, 'xsect_v.npy']))

    def xsect_triangles(self):
        """cross section triangles"""
        return npload(sep.join([self.xsect_surface_dir, 'xsect_t.npy']))

    def xsect_data(self):
        """cross section raw data"""
        raw_data = []
        for npyfile in glob(sep.join([self.xsect_surface_dir, '*.npy'])):
            if npyfile.endswith('_v.npy') or npyfile.endswith('_t.npy'):
                continue
            raw_data += [dict(
                location='CC',
                data=DataArray(
                    title=npyfile.split(sep)[-1][:-4],
                    array=npload(npyfile)
                )
            )]
        return raw_data

    def xsect_surface(self):
        """Steno3D surface resource of cross sections"""
        return Surface(
            project=self.project,
            mesh=Mesh2D(
                vertices=self.xsect_vertices,
                triangles=self.xsect_triangles
            ),
            data=self.xsect_data,
            title='Cross-Sections'
        )

    # Lithology Volume:
    def lith_volume_dir(self):
        """directory for lithology volume"""
        return sep.join([self.wolfpass_dir, 'lith_volume'])

    def lith_tensor(self):
        """(h1, h2, h3, x0) for lith volume"""
        return (
            npload(sep.join([self.lith_volume_dir, 'vol_h1.npy'])),
            npload(sep.join([self.lith_volume_dir, 'vol_h2.npy'])),
            npload(sep.join([self.lith_volume_dir, 'vol_h3.npy'])),
            npload(sep.join([self.lith_volume_dir, 'vol_x0.npy']))
        )

    def lith_data(self):
        """raw data for lith volume"""
        raw_data = []
        for npyfile in glob(sep.join([self.lith_volume_dir, '*.npy'])):
            if npyfile.split(sep)[-1].startswith('vol_'):
                continue
            raw_data += [dict(
                location='CC',
                data=DataArray(
                    title=npyfile.split(sep)[-1][:-4],
                    array=npload(npyfile)
                )
            )]
        return raw_data

    def lith_volume(self):
        """Steno3D volume resource of lithology"""
        return Volume(
            project=self.project,
            mesh=Mesh3DGrid(
                h1=self.lith_tensor[0],
                h2=self.lith_tensor[1],
                h3=self.lith_tensor[2],
                x0=self.lith_tensor[3]
            ),
            data=self.lith_data,
            title='Lithology Volume'
        )

    def project(self):
        """empty Steno3D project"""
        return Project(
            title='Wolf Pass'
        )

    def get_resources(self):
        """get a copy of the Wolfpass resources"""
        res = (self.collar_points, self.borehole_lines,)
        res += self.cu_surfaces
        res += self.lith_surfaces
        res += (self.topo_surface, self.xsect_surface, self.lith_volume)
        return res

    def get_project(self):
        """get a copy of the Wolfpass project"""
        return self.get_resources()[0].project[0]



"""resource.py contains the base resource classes that user-created
resources depend on in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties

from .client import pause
from .content import UserContent


class BaseResource(UserContent):
    """Base class for all resources that are added to projects and
    uploaded to steno3d
    """
    meta = properties.Object('Meta data in JSON form.', default={})

    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        return True


class CompositeResource(BaseResource):
    """A composite resource that stores references to lower-level
    objects, and may also grant access to them through ACL delegation.
    """
    project = properties.Pointer(
        'Project',
        ptype='Project',
        required=True,
        repeated=True
    )
    _children = {
        'mesh': None,
        'data': 'data',
        'textures': None,
    }

    def __init__(self, project=None, **kwargs):
        if project is None:
            raise TypeError('Resource must be constructed with its '
                            'containing project(s)')
        super(CompositeResource, self).__init__(**kwargs)
        self.project = project

    @properties.validator
    def validate(self):
        for proj in self.project:
            if self not in proj.resources:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that projects contain all the '
                                 'resources that point to them.')

    def upload(self, *args, **kwargs):
        """Upload the resource through its containing project(s)"""
        # self.validate()
        for proj in self.project:
            proj.upload(*args, **kwargs)

    def _upload(self, *args, **kwargs):
        """Perform the upload and any follow-up tasks"""
        pause()
        res = super(CompositeResource, self)._upload(*args, **kwargs)
        pause()
        resp = self.add_recursive_view_permissions()
        if not resp:
            raise Exception(('Could not add delegation permissions '
                             'for one or more child object(s)'))
        return res

    def _on_property_change(self, name, pre, post):
        if name == 'project':
            if pre is None:
                pre = []
            if post is None:
                post = []
            for proj in post:
                if proj not in pre:
                    proj.resources += [self]
            for proj in pre:
                if proj not in post:
                    proj.resources = [r for r in proj.resources
                                      if r is not self]
            if len(set(post)) != len(post):
                post_post = []
                for p in post:
                    if p not in post_post:
                        post_post += [p]
                self.project = post_post

    def add_recursive_view_permissions(self, accessor=None):
        """Add recursive view permissions to child objects for
        this class or another accessor (e.g., a project)
        """
        if accessor is None:
            accessor = self
        acl_responses = []
        for key in self._children:
            subkey = self._children[key]
            child = getattr(self, key, None)
            if child:
                if isinstance(child, list):
                    for item in child:
                        target = getattr(item, subkey) if subkey else item
                        pause()
                        acl_responses.append(
                            target.add_view_permissions(accessor)
                        )
                else:
                    target = getattr(child, subkey) if subkey else child
                    pause()
                    acl_responses.append(target.add_view_permissions(accessor))
        return all(acl_responses)


class BaseMesh(BaseResource):
    """Base class for all mesh resources. These are contained within
    each composite resources and define its structure
    """
    pass


class BaseData(BaseResource):
    """Base class for all data resources. These can be contained within
    each composite resource and define data corresponding to the mesh
    """
    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/data/{className}'.format(
                className=cls._resource_class)
        return cls.__model_api_location


class BaseTexture2D(BaseResource):
    """Base class for all texture resources. These can be contained
    within some composite resources and define data in space that gets
    mapped to the mesh.
    """
    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for texture resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/texture2d/{className}'.format(
                className=cls._resource_class)
        return cls.__model_api_location



"""data.py contains resource data structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties

from .base import BaseData


class DataArray(BaseData):
    """Data array with unique values at every point in the mesh"""
    _resource_class = 'array'
    array = properties.Array(
        'Data, unique values at every point in the mesh',
        shape=('*',),
        dtype=(float, int)
    )

    def __init__(self, array=None, **kwargs):
        super(DataArray, self).__init__(**kwargs)
        if array is not None:
            self.array = array

    def validate(self):
        """Check if content is built correctly"""
        return True

    @property
    def _files(self):
        """Get the file information"""
        files = {
            'array': self._properties['array'].serialize(self.array),
        }
        return files

__all__ = ['DataArray']



"""line.py contains the Line composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
from numpy import max as npmax
from numpy import min as npmin
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh1DOptions(MeshOptions):
    pass


class _LineOptions(ColorOptions):
    pass


class Mesh1D(BaseMesh):
    """Contains spatial information of a 1D line set"""
    vertices = properties.Array(
        'Mesh vertices',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    segments = properties.Array(
        'Segment endpoint indices',
        shape=('*', 2),
        dtype=int,
        required=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_Mesh1DOptions
    )

    def __init__(self, *args, **kwargs):
        super(Mesh1D, self).__init__(*args, **kwargs)
        self._nN = None
        self._nC = None

    @property
    def nN(self):
        """ get number of nodes """
        if getattr(self, '_nN', None) is None:
            self._nN = len(self.vertices)
        return self._nN

    @property
    def nC(self):
        """ get number of cells """
        if getattr(self, '_nC', None) is None:
            self._nC = len(self.segments)
        return self._nC

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if npmin(self.segments) < 0:
            raise ValueError('Segments may only have positive integers')
        if npmax(self.segments) >= len(self.vertices):
            raise ValueError('Segments expects more vertices than provided')
        return True

    @property
    def _files(self):
        """Get the file information"""
        files = {
            'vertices': self._properties['vertices'].serialize(self.vertices),
            'segments': self._properties['segments'].serialize(self.segments),
        }
        return files


class _LineBinder(properties.PropertyClass):
    """Contains the data on a 1D line set with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('LINE', 'FACE', 'CELLCENTER', 'EDGE', 'SEGMENT'),
            'N': ('VERTEX', 'NODE', 'ENDPOINT')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Line(CompositeResource):
    """Contains all the information about a 1D line set"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh1D,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_LineBinder,
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_LineOptions
    )

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'line.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Line, self).validate()
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""
        datadict = {
            'mesh': dumps({
                'uid': self.mesh.json_after_upload['longUid'],
            }),
            'data': dumps([
                {
                    'location': d.location,
                    'uid': d.data.json_after_upload['longUid'],
                } for d in self.data
            ]),
        }
        return datadict


__all__ = ['Line', 'Mesh1D']



"""mics.py contains miscellaneous files and components to supplement
Steno3D examples
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import sep

from .base import BaseExample


class Images(BaseExample):
    """Class containing miscellaneous image files."""

    def metal(self):
        """metal texture png"""
        return sep.join([self.asset_dir, 'basic', 'metal.png'])

    def wood(self):
        """wood texture png"""
        return sep.join([self.asset_dir, 'basic', 'woodplanks.png'])

try:
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from os.path import sep

from json import loads
from numpy import array
from numpy import mean

from .base import BaseExample
from ..data import DataArray
from ..project import Project
from ..surface import Mesh2D
from ..surface import Surface
from ..texture import Texture2DImage


class Topography(BaseExample):

    def file_base(self):
        """topography file base"""
        return sep.join([self.asset_dir, 'basic', 'topography'])

    def data(self):
        """topography data from json"""
        with open(self.file_base + '.json', 'r') as f:
            data = loads(f.read())
        return data

    def vertices(self):
        """n x 3 topography vertices"""
        return array(self.data['vertices'])

    def triangles(self):
        """n x 3 topography triangles"""
        return array(self.data['triangles'])

    def texture(self):
        """topography surface image texture"""
        return Texture2DImage(
            O=[443200, 491750, 0],
            U=[4425, 0, 0],
            V=[0, 3690, 0],
            image=self.file_base + '.png'
        )

    def vertex_elevation(self):
        """z-coordinate of vertex"""
        return self.vertices[:, 2]

    def vertex_elevation_data(self):
        """elevation as Steno3D data"""
        return DataArray(
            array=self.vertex_elevation,
            title='Elevation, vertices'
        )

    def face_elevation(self):
        """average z-coordinate of face"""
        return mean(self.vertex_elevation[self.triangles], axis=1)

    def face_elevation_data(self):
        """elevation as Steno3D data"""
        return DataArray(
            array=self.face_elevation,
            title='Elevation, faces'
        )

    def surface(self):
        """ground surface with topo data and surface imagery"""
        return Surface(
            project=self.project,
            mesh=Mesh2D(
                vertices=self.vertices,
                triangles=self.triangles
            ),
            data=[
                dict(
                    location='N',
                    data=self.vertex_elevation_data
                ),
                dict(
                    location='CC',
                    data=self.face_elevation_data
                )
            ],
            textures=self.texture,
            title='Topography Surface',
            description=('This surface has face and vertex elevation '
                         'data as well as surface imagery')
        )

    def project(self):
        """empty Steno3D project"""
        return Project(
            title='Topography and Surface Imagery'
        )

    def get_resources(self):
        """get a copy of the topography resources

        tuple(topography surface,)
        """
        return (self.surface,)

    def get_project(self):
        """get a copy of the topography project"""
        return self.get_resources()[0].project[0]



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from properties.base import Pointer

from . import parsers
from . import client
from .project import *
from .data import *
from .line import *
from .point import *
from .surface import *
from .texture import *
from .volume import *

__version__ = u'0.0.2'
__author__ = u'3point Science, Inc.'
__license__ = u'MIT'
__copyright__ = u'Copyright 2016 3point Science, Inc.'

login = client.Comms.login
logout = client.Comms.logout
Pointer.resolve()

try:
    del Pointer
    del project, data, line, point, surface, texture, volume
    del base, client, content, options
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



"""clinet.py contains the functionality to link the python steno3d
client with the steno3d website
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals
from builtins import str
from builtins import input

import keyring
import requests
from six import string_types
from six.moves.urllib.parse import urlparse
from time import sleep


PRODUCTION_BASE_URL = 'https://steno3d.com/'
API_SUBPATH = 'api/'
ACCEPTABLE_ACL_ERRORS = set(['repeated users', 'to delete'])
SLEEP_TIME = 0.1

WELCOME_HEADER = """

Welcome to the Python client library for Steno3D!

"""

DEVKEY_PROMPT = "If you have a Steno3D developer key, please enter it here > "

WELCOME_MESSAGE = """

If you do not have a Steno3D developer key, you need to request
one from the Steno3D website in order to access the API. Please
log in to the application (if necessary) and request a new key.

{base_url}settings/api

If you are not yet signed up, you can do that here:

{base_url}signup

When you are ready, please enter the key above, or reproduce this
prompt by calling steno3d.login().

"""

LOGIN_FAILED = """

Oh no! We could not log you in.

The API developer key that you provided could not be validated. You could:

1) Clear your keychain with `steno3d.logout()` and try again
2) Restart your Python kernel and try again
3) Check that you have the correct API key
4) Ask for <help@steno3d.com>
5) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

NOT_CONNECTED = """

Oh no! We could not connect to the Steno3D server.

Please ensure that you are:

1) connected to the Internet!
2) Can connect to Steno3D on  https://steno3d.com

If the issue persists please:

1) Ask for <help@steno3d.com>
2) Open an issue https://github.com/3ptscience/steno3dpy/issues

"""

BAD_API_KEY = """

The developer key should be 36 characters.
Please checkout the website for your key:

{base_url}settings/api

"""


class _Comms(object):
    """Comms controls the interaction between the python client and the
    steno3d website.
    """

    def __init__(self):
        self._user = None
        self._base_url = PRODUCTION_BASE_URL
        self._hard_devel_key = None

    @property
    def host(self):
        """hostname of url"""
        parseresult = urlparse(self.base_url)
        return parseresult.hostname

    @property
    def url(self):
        """url endpoint for uploading"""
        return self.base_url + API_SUBPATH

    @property
    def base_url(self):
        """base url endpoint for uploading"""
        return getattr(self, '_base_url', PRODUCTION_BASE_URL)

    @base_url.setter
    def base_url(self, value):
        assert isinstance(value, string_types), \
            'Endpoint path must be a string'
        # Patch '/' onto bare URL endpoints
        if not value[-1] == '/':
            value += '/'
        # Check for HTTPS
        parsed = urlparse(value)
        if '.com' in parsed.hostname and parsed.scheme != 'https':
            raise Exception('Live endpoints require HTTPS.')

        self._base_url = value

    @property
    def devel_key(self):
        """developer key acquired from steno3d.com"""
        self._user = None
        if getattr(self, '_hard_devel_key', None) is not None:
            return self._hard_devel_key

        key = keyring.get_password('steno3d', self.host)
        if key in {None, 'None'}:
            return None
        return str(key)

    @devel_key.setter
    def devel_key(self, value):
        if value is None:
            del self.devel_key
        else:
            keyring.set_password('steno3d', self.host, value)

    @devel_key.deleter
    def devel_key(self):
        try:
            keyring.delete_password('steno3d', self.host)
        except keyring.errors.PasswordDeleteError:
            # Happens when the key does not exist
            pass

    def login(self, devel_key=None, skip_keychain=False, endpoint=None):
        """Login to steno3d.com to allow uploading resources"""
        if endpoint is not None:
            self.base_url = str(endpoint)

        if devel_key is not None:
            if skip_keychain:
                self._hard_devel_key = devel_key
            else:
                self.devel_key = devel_key

        if self.devel_key is None:
            print(WELCOME_MESSAGE.format(base_url=self.base_url))
            try:
                devel_key = raw_input(WELCOME_HEADER + DEVKEY_PROMPT)
            except NameError:
                devel_key = input(WELCOME_HEADER + DEVKEY_PROMPT)
            if len(devel_key) is not 36:
                print(BAD_API_KEY.format(base_url=self.base_url))
                try:
                    devel_key = raw_input(DEVKEY_PROMPT)
                except NameError:
                    devel_key = input(DEVKEY_PROMPT)
            self.devel_key = devel_key

        if getattr(self, '_user', None) is None:
            try:
                resp = requests.get(
                    self.url + 'me',
                    headers={'sshKey': self.devel_key}
                )
            except requests.ConnectionError:
                raise Exception(NOT_CONNECTED)
            if resp.status_code is not 200:
                self.devel_key = None
                raise Exception(LOGIN_FAILED)
            self._user = resp.json()

        username = self._user['uid']
        print(
            'Welcome to Steno3D! You are logged in as @{name}'.format(
                name=username
            )
        )

    def logout(self):
        """Logout current user and remove api key from keyring"""
        self.devel_key = None


Comms = _Comms()


def pause():
    """Brief pause on localhost to simulate network delay"""
    if 'localhost' in Comms.url:
        sleep(SLEEP_TIME)


def post(url, data=None, files=None):
    """Send data and files to the steno3d online endpoint"""
    data = {} if data is None else data
    filedict = {}
    for filename in files:
        if hasattr(files[filename], 'dtype'):
            filedict[filename] = files[filename].file
            filedict[filename + 'Type'] = files[filename].dtype
        else:
            filedict[filename] = files[filename]
    req = requests.post(
        Comms.url + url,
        data=data,
        files=filedict,
        headers={'sshKey': Comms.devel_key}
    )
    for key in files:
        files[key].file.close()
    return req


def add_acl(resource, accessor, role):
    """Add permissions to an object's ACL"""
    long_uid = (
        accessor if isinstance(accessor, string_types)
        else accessor.long_uid
    )
    data = {
        'user': long_uid,
        'role': role,
    }
    pause()
    response = requests.post(
        resource.url_api + '/access',
        data=data,
        headers={'sshKey': Comms.devel_key}
    )
    return response


def add_view_permissions(resource, accessor):
    """Add view permissions to an object"""
    return add_acl(resource, accessor, 'view')


def add_edit_permissions(resource, accessor):
    """Add edit permissions to an object"""
    return add_acl(resource, accessor, 'edit')


def remove_permissions(resource, accessor):
    """Remove permissions from an object"""
    long_uid = (
        accessor if isinstance(accessor, string_types)
        else accessor.long_uid
    )
    response = requests.delete(
        resource.url_api + '/access/{0}'.format(long_uid),
        headers={'sshKey': Comms.devel_key}
    )
    return response


def check_acl_response(resp):
    """Check for 'acceptable' errors"""
    if resp.status_code == 200:
        return True
    respjson = resp.json()
    return 'reason' in respjson and \
        any(telltale in respjson['reason']
            for telltale in ACCEPTABLE_ACL_ERRORS)


def plot(url):
    """Return an IFrame plot"""
    from IPython.display import IFrame
    return IFrame(url, width='100%', height=500)



"""parsers.py creates the groundwork for parsers to be plugged in to
steno3d. When a parser is imported, it gets added to the steno3d.parsers
namespace.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from future.utils import with_metaclass as _with_metaclass
from properties import PropertyClass as _PropertyClass
from six import string_types as _string_types


class _ParserMetaClass(type):
    """Metaclass to ensure Parser classes fit the requried format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != '_BaseParser':
            assert 'extensions' in attrs, \
                "You must say the extensions that are in the {name} parser. " \
                "e.g. ('obj',)".format(name=name)
            assert isinstance(attrs['extensions'], tuple), \
                'The `extensions` in the {name} parser must be a tuple of ' \
                'known extensions.'.format(name=name)
            for ext in attrs['extensions']:
                assert isinstance(ext, _string_types), \
                    'Extensions in the {name} parser must be ' \
                    'strings'.format(name=name)
        new_class = super(_ParserMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class _BaseParserMetaClass(_ParserMetaClass,
                           _PropertyClass.__class__):
    """Augmented metaclass for parsers, inherits from properties
    metaclass
    """


class _BaseParser(_with_metaclass(_BaseParserMetaClass,
                                  _PropertyClass)):
    """Base class for Steno3D parser objects"""


class _DirectorMetaClass(type):
    """Metaclass to ensure Director classes fit the requried format and
    get added to the steno3d.parsers namespace
    """

    def __new__(mcs, name, bases, attrs):
        if name != 'AllParsers':
            assert name.startswith('AllParsers_'), \
                "Parser classes that inherit 'AllParsers' such as {name} " \
                "must have names that start with " \
                "'AllParsers_'".format(name=name)
            assert 'extensions' in attrs, \
                "{name} must contain a dictionary of extensions and " \
                "parsers".format(name=name)
            assert isinstance(attrs['extensions'], dict), \
                "{name} extensions must be a dictionary of extensions and " \
                "supporting parser".format(name=name)
            for ext in attrs['extensions']:
                assert isinstance(ext, _string_types), \
                    'Extensions in {name} must be strings'.format(name=name)
                assert issubclass(type(attrs['extensions'][ext]),
                                  _BaseParserMetaClass), \
                    'Extensions in {name} must direct to a ' \
                    'Parser class'.format(name=name)
        new_class = super(_DirectorMetaClass, mcs).__new__(
            mcs, name, bases, attrs
        )
        globals()[name] = new_class
        return new_class


class _BaseDirectorMetaClass(_DirectorMetaClass,
                             _PropertyClass.__class__):
    """Augmented metaclass for parser Directors, inherits from
    properties metaclass
    """


class AllParsers(_with_metaclass(_BaseDirectorMetaClass,
                                 _PropertyClass)):
    """Base class for Steno3D parser objects that parse all
    available file types
    """

    def __new__(cls, filename, **kwargs):
        if getattr(cls, 'extensions', None) is None:
            cls.extensions = dict()
            parser_keys = [
                k for k in globals()
                if (
                    k != '_BaseParser' and
                    issubclass(type(globals()[k]), _BaseParserMetaClass)
                )
            ]
            for k in parser_keys:
                for ext in globals()[k].extensions:
                    if ext not in cls.extensions:
                        cls.extensions[ext] = globals()[k]
                    elif issubclass(type(cls.extensions[ext]),
                                    _BaseParserMetaClass):
                        cls.extensions[ext] = (cls.extensions[ext].__name__ +
                                               ', ' + globals()[k].__name__)
                    else:
                        cls.extensions[ext] = (cls.extensions[ext] +
                                               ', ' + globals()[k].__name__)

        for ext in cls.extensions:
            if filename.split('.')[-1] == ext:
                if not issubclass(type(cls.extensions[ext]),
                                  _BaseParserMetaClass):
                    raise ValueError(
                        '{ext}: file type supported by more than one parser. '
                        'Please specify one of ({parsers})'.format(
                            ext=ext,
                            parsers=cls.extensions[ext]
                        )
                    )
                return cls.extensions[ext](filename, **kwargs)

        raise ValueError(
            '{bad}: unsupported file extensions. Must be in ({ok})'.format(
                bad=filename.split('.')[-1],
                ok=', '.join(list(cls.extensions))
            )
        )

try:
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from .teapot import Teapot
from .topography import Topography
from .wolfpass import Wolfpass
from . import misc

try:
    del teapot, topography, wolfpass
    del base
    del absolute_import, division, print_function, unicode_literals
except NameError:
    # Error cleaning namespace
    pass



"""content.py contains the base class for everything users can create
in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from pprint import pformat
import properties

from . import client


class UserContent(properties.PropertyClass):
    """Base class for everything user creates and uploads in steno3d"""
    title = properties.String(
        'Title of the model.',
        default=''
    )
    description = properties.String(
        'Description of the model.',
        default=''
    )
    opts = None

    def __init__(self, **kwargs):
        super(UserContent, self).__init__(**kwargs)
        self._upload_data = None

    @properties.classproperty
    @classmethod
    def _resource_class(cls):
        """name of the class of resource"""
        if getattr(cls, '__resource_class', None) is None:
            cls.__resource_class = cls.__name__.lower()
        return cls.__resource_class

    @properties.classproperty
    @classmethod
    def _model_api_location(cls):
        """api destination for resource"""
        if getattr(cls, '__model_api_location', None) is None:
            cls.__model_api_location = 'resource/{className}'.format(
                className=cls._resource_class
            )
        return cls.__model_api_location

    # @property
    # def _dirty(self):
    #     """Set containing unsaved properties"""
    #     if not hasattr(self, '__dirty'):
    #         self.__dirty = set()
    #     return self.__dirty

    @classmethod
    def _url_api_from_uid(cls, uid):
        """Get api location object from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=client.Comms.url,
            mapi=cls._model_api_location,
            uid=uid)
        return url

    @classmethod
    def _url_view_from_uid(cls, uid):
        """Get full url from a uid"""
        url = '{base}{mapi}/{uid}'.format(
            base=client.Comms.base_url,
            mapi=cls._model_api_location,
            uid=uid)
        return url

    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        return True

    @property
    def _contentdata(self):
        """Get basic data about the content"""
        datadict = {
            'title': self.title,
            'description': self.description
        }
        if getattr(self, 'opts', None) is not None:
            datadict['meta'] = self.opts.as_json()
        return datadict

    @property
    def _reqdata(self):
        """Get the required longUids"""
        return {}

    @property
    def _files(self):
        """Get the file information"""
        return {}

    @property
    def _datadict(self):
        """Combine metadata and requirement information for upload"""
        datadict = self._contentdata.copy()
        datadict.update(self._reqdata)
        return datadict

    @property
    def long_uid(self):
        """longUid of uploaded data"""
        if getattr(self, '_upload_data', None) is not None:
            return self._upload_data['longUid']
        else:
            return None

    @property
    def uid(self):
        """uid of uploaded data"""
        if getattr(self, '_upload_data', None) is not None:
            return self._upload_data['uid']
        else:
            return None

    @property
    def url_api(self):
        """api location of uploaded data"""
        uid = self.uid
        if uid is not None:
            return self._url_api_from_uid(uid)
        else:
            return None

    @property
    def url_view(self):
        """full url of uploaded data"""
        uid = self.uid
        if uid is not None:
            return self._url_view_from_uid(uid)
        else:
            return None

    def _upload(self, force=True, pause=False):
        """upload the UserContent to the endpoint selected at login"""
        if getattr(self, '_upload_data', None) is not None and not force:
            return self.url_view

        assert self.validate()

        datadict = self._datadict
        files = self._files
        client.pause()
        req = client.post(
            self._model_api_location,
            data=datadict if datadict else tuple(),
            files=files if files else tuple(),
        )

        if isinstance(req, list):
            for rq in req:
                if rq.status_code != 200:
                    try:
                        resp = pformat(rq.json())
                    except ValueError:
                        resp = rq

                    raise Exception(
                        'Upload failed: {location}'.format(
                            location=self._model_api_location,
                        ) +
                        '\ndata: {datadict}\nfiles: {filedict}'.format(
                            datadict=pformat(datadict),
                            filedict=pformat(files),
                        ) +
                        '\nresponse: {response}'.format(
                            response=resp,
                        )
                    )
            self._upload_data = [rq.json() for rq in req]
        else:
            if req.status_code != 200:
                try:
                    resp = pformat(req.json())
                except ValueError:
                    resp = req
                raise Exception(
                    'Upload failed: {location}'.format(
                        location=self._model_api_location,
                    ) +
                    '\ndata: {datadict}\nfiles: {filedict}'.format(
                        datadict=pformat(datadict),
                        filedict=pformat(files),
                    ) +
                    '\nresponse: {response}'.format(
                        response=resp,
                    )
                )
            self._upload_data = req.json()

        if pause:
            client.pause()

        return self.url_view

    def plot(self):
        """Display the 3D representation of the content"""
        if getattr(self, '_upload_data', None) is None:
            raise Exception('Plotting failed: resource not uploaded - '
                            'please upload() first.')
        return client.plot(self.url_view)

    @property
    def json_after_upload(self):
        """Return a JSON representation of the object,
        uploading if necessary.
        """
        if getattr(self, '_upload_data', None) is None:
            self._upload()
        return self.json

    @property
    def json(self):
        """Return a JSON representation of the object"""
        return getattr(self, '_upload_data', None)

    def grant_access(self, userid):
        """Grant a user access to this object"""
        if userid.isalnum():
            userid = 'User:' + userid
        resp = client.add_edit_permissions(self, userid)
        return client.check_acl_response(resp)

    def revoke_access(self, userid):
        """Revoke a user's access to this object"""
        if userid.isalnum():
            userid = 'User:' + userid
        resp = client.remove_permissions(self, userid)
        return client.check_acl_response(resp)

    def add_view_permissions(self, accessor):
        """Add view permissions to this object for an accessor"""
        client.pause()
        resp = client.add_view_permissions(self, accessor)
        if resp.status_code == 200:
            return True
        return client.check_acl_response(resp)



"""volume.py contains the Volume composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh3DOptions(MeshOptions):
    pass


class _VolumeOptions(ColorOptions):
    pass


class Mesh3DGrid(BaseMesh):
    """Contains spatial information of a 3D grid volume."""

    h1 = properties.Array(
        'Tensor cell widths, x-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h2 = properties.Array(
        'Tensor cell widths, y-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    h3 = properties.Array(
        'Tensor cell widths, z-direction',
        shape=('*',),
        dtype=float,
        required=True
    )
    x0 = properties.Vector(
        'Origin vector',
        default=[0, 0, 0]
    )
    opts = properties.Pointer(
        'Mesh3D Options',
        ptype=_Mesh3DOptions
    )

    def __init__(self, *args, **kwargs):
        super(Mesh3DGrid, self).__init__(*args, **kwargs)
        self._nN = None
        self._nC = None

    @property
    def nN(self):
        """ get number of nodes """
        if getattr(self, '_nN', None) is None:
            self._nN = (
                (len(self.h1)+1) * (len(self.h2)+1) * (len(self.h3)+1)
            )
        return self._nN

    @property
    def nC(self):
        """ get number of cells """
        if getattr(self, '_nC', None) is None:
            self._nC = len(self.h1) * len(self.h2) * len(self.h3)
        return self._nC

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.x0.nV != 1:
            raise ValueError('Origin x0 must be only one vector')
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""
        datadict = {
            'tensors': dumps(dict(
                h1=self.h1.tolist(),
                h2=self.h2.tolist(),
                h3=self.h3.tolist(),
            )),
            'OUVZ': dumps(dict(
                O=self.x0.tolist()[0],
                U=[self.h1.sum(), 0, 0],
                V=[0, self.h2.sum(), 0],
                Z=[0, 0, self.h3.sum()],
            )),
        }
        return datadict


class _VolumeBinder(properties.PropertyClass):
    """Contains the data on a 3D volume with location information"""
    location = properties.String(
        'Location of the data on mesh',
        required=True,
        choices={
            'CC': ('CELLCENTER'),
            'N': ('NODE', 'VERTEX', 'CORNER')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Volume(CompositeResource):
    """Contains all the information about a 3D volume"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh3DGrid,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_VolumeBinder,
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_VolumeOptions
    )

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location in ('N', 'CC')
            valid_length = (
                self.mesh.nC if dat.location == 'CC'
                else self.mesh.nN
            )
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'volume.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Volume, self).validate()
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""
        datadict = {
            'mesh': dumps({
                'uid': self.mesh.json_after_upload['longUid'],
            }),
            'data': dumps([
                {
                    'location': d.location,
                    'uid': d.data.json_after_upload['longUid'],
                } for d in self.data
            ]),
        }
        return datadict


__all__ = ['Volume', 'Mesh3DGrid']



"""resource.py contains the base resource classes that user-created
resources depend on in steno3d
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import properties

from .client import pause
from .content import UserContent


class Project(UserContent):
    """Steno3D top-level project"""
    _model_api_location = 'project/steno3d'

    resources = properties.Pointer(
        'Project Resources',
        ptype='CompositeResource',
        repeated=True
    )

    public = properties.Bool(
        'Public visibility of project',
        default=False
    )

    def upload(self, *args, **kwargs):
        """Upload the project"""
        print(self._upload(*args, **kwargs))

    @properties.validator
    def validate(self):
        """Check if project resource pointers are correct"""
        for res in self.resources:
            if self not in res.project:
                raise ValueError('Project/resource pointers misaligned: '
                                 'Ensure that resources point to containing '
                                 'project.')
        return True

    def _on_property_change(self, name, pre, post):
        if name == 'resources':
            if pre is None:
                pre = []
            if post is None:
                post = []
            for res in post:
                if res not in pre:
                    res.project += [self]
            for res in pre:
                if res not in post:
                    res.project = [p for p in res.project
                                   if p is not self]
            if len(set(post)) != len(post):
                post_post = []
                for r in post:
                    if r not in post_post:
                        post_post += [r]
                self.resources = post_post

    @property
    def _contentdata(self):
        """Get basic data about the content"""
        datadict = {
            'title': self.title,
            'description': self.description,
            'public': self.public
        }
        return datadict

    @property
    def _reqdata(self):
        """get the required longUids"""
        resource_reqs = (r.json_after_upload for r in self.resources)
        datadict = {
            'resourceUids': ','.join((r['longUid'] for r in resource_reqs)),
        }
        return datadict

    def _upload(self, *args, **kwargs):
        """Perform the upload and any follow-up tasks"""
        res = super(Project, self)._upload(*args, **kwargs)
        pause()
        acl_responses = []
        for resource in self.resources:
            acl_responses.append(resource.add_view_permissions(self))
            resource.add_recursive_view_permissions(self)
        if all(acl_responses):
            return res
        raise Exception(
            'Could not add delegation permissions for one or more '
            'child object(s)'
        )


__all__ = ['Project']



"""point.py contains the Point composite resource for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
import properties

from .base import BaseMesh
from .base import CompositeResource
from .options import ColorOptions
from .options import MeshOptions


class _Mesh0DOptions(MeshOptions):
    pass


class _PointOptions(ColorOptions):
    pass


class Mesh0D(BaseMesh):
    """Contains spatial information of a 0D point cloud."""
    vertices = properties.Array(
        'Point locations',
        shape=('*', 3),
        dtype=float,
        required=True
    )
    opts = properties.Pointer(
        'Mesh0D Options',
        ptype=_Mesh0DOptions
    )

    def __init__(self, *args, **kwargs):
        super(Mesh0D, self).__init__(*args, **kwargs)
        self._nN = None

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        return True

    @property
    def nN(self):
        """ get number of nodes """
        if getattr(self, '_nN', None) is None:
            self._nN = len(self.vertices)
        return self._nN

    @property
    def _files(self):
        """Get the file information"""
        files = {
            'vertices': self._properties['vertices'].serialize(self.vertices),
        }
        return files


class _PointBinder(properties.PropertyClass):
    """Contains the data on a 0D point cloud"""
    location = properties.String(
        'Location of the data on mesh',
        default='N',
        required=True,
        choices={
            'N': ('NODE', 'CELLCENTER', 'CC', 'VERTEX')
        }
    )
    data = properties.Pointer(
        'Data',
        ptype='DataArray',
        required=True
    )


class Point(CompositeResource):
    """Contains all the information about a 0D point cloud"""
    mesh = properties.Pointer(
        'Mesh',
        ptype=Mesh0D,
        required=True
    )
    data = properties.Pointer(
        'Data',
        ptype=_PointBinder,
        repeated=True
    )
    textures = properties.Pointer(
        'Textures',
        ptype='Texture2DImage',
        repeated=True
    )
    opts = properties.Pointer(
        'Options',
        ptype=_PointOptions
    )

    @properties.validator
    def validate(self):
        """Check if resource is built correctly"""
        for ii, dat in enumerate(self.data):
            assert dat.location == 'N'
            valid_length = self.mesh.nN
            if len(dat.data.array) != valid_length:
                raise ValueError(
                    'point.data[{index}] length {datalen} does not match '
                    '{loc} length {meshlen}'.format(
                        index=ii,
                        datalen=len(dat.data.array),
                        loc=dat.location,
                        meshlen=valid_length
                    )
                )
        super(Point, self).validate()
        return True

    @property
    def _reqdata(self):
        """Get the required longUids"""
        datadict = {
            'mesh': dumps({
                'uid': self.mesh.json_after_upload['longUid'],
            }),
            'data': dumps([
                {
                    'location': d.location,
                    'uid': d.data.json_after_upload['longUid'],
                } for d in self.data
            ]),
            'textures': dumps([
                {
                    'uid': t.json_after_upload['longUid'],
                } for t in self.textures
            ]),
        }
        return datadict


__all__ = ['Point', 'Mesh0D']



"""texture.py contains the texture resource structures"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from collections import namedtuple
from json import dumps
import properties

from .base import BaseTexture2D


FileProp = namedtuple('FileProp', ['file', 'dtype'])


class Texture2DImage(BaseTexture2D):
    """Contains an image that can be mapped to a 2D surface"""

    _resource_class = 'image'

    O = properties.Vector(
        'Origin of the texture',
        required=True
    )
    U = properties.Vector(
        'U axis of the texture',
        required=True
    )
    V = properties.Vector(
        'V axis of the texture',
        required=True
    )
    image = properties.Image(
        'Image file',
        required=True
    )

    @properties.validator
    def validate(self):
        """Check if mesh content is built correctly"""
        if self.O.nV != 1 or self.U.nV != 1 or self.V.nV != 1:
            raise ValueError('O, U, and V must each be only one vector')
        return True

    @property
    def _files(self):
        """Get the file information"""
        self.image.seek(0)
        files = {
            'image': FileProp(self.image, 'png')
        }
        return files

    @property
    def _reqdata(self):
        datadict = {
            'OUV': dumps(dict(
                O=[self.O[0][0], self.O[0][1], self.O[0][2]],
                U=[self.U[0][0], self.U[0][1], self.U[0][2]],
                V=[self.V[0][0], self.V[0][1], self.V[0][2]],
            )),
        }
        return datadict

    def _repr_png_(self):
        """For IPython display"""
        if self.image is None:
            return None
        self.image.seek(0)
        return self.image.read()


__all__ = ['Texture2DImage']



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import subprocess
import unittest
import os


class TestDoc(unittest.TestCase):

    @property
    def docs_dir(self):
        dirname, filename = os.path.split(os.path.abspath(__file__))
        return os.path.sep.join(dirname.split(os.path.sep)[:-1] + ['docs'])

    def setUp(self):
        self.build_dir = os.path.sep.join(
            self.docs_dir.split(os.path.sep) + ['_build']
        )
        if not os.path.isdir(self.build_dir):
            subprocess.call(['mkdir', '{0}'.format(self.build_dir)])

        self.doctrees_dir = os.path.sep.join(
            self.build_dir.split(os.path.sep) + ['doctrees']
        )
        if not os.path.isdir(self.doctrees_dir):
            subprocess.call(['mkdir', '{0}'.format(self.doctrees_dir)])

        self.html_dir = os.path.sep.join(
            self.build_dir.split(os.path.sep) + ['html']
        )
        if not os.path.isdir(self.html_dir):
            subprocess.call(['mkdir', '{0}'.format(self.html_dir)])

    def test_html(self):
        check = subprocess.call(
            ['sphinx-build', '-nW', '-b', 'html', '-d',
             '{}'.format(self.doctrees_dir),
             '{}'.format(self.docs_dir),
             '{}'.format(self.html_dir)]
        )
        assert check == 0

    def test_linkcheck(self):
        check = subprocess.call(
            ['sphinx-build', '-nW', '-b', 'linkcheck', '-d',
             '{}'.format(self.doctrees_dir),
             '{}'.format(self.docs_dir),
             '{}'.format(self.build_dir)]
        )
        assert check == 0

if __name__ == '__main__':
    unittest.main()



"""options.py defines the base option classes for steno3d"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

from json import dumps
import properties


class Options(properties.PropertyClass):
    """Generic options for all steno3d resources"""

    def as_json(self):
        """returns json representation of options"""
        opts_json = {}
        for key in self._properties:
            opts_json[key] = getattr(self, key)
        return dumps(opts_json)

    @properties.validator
    def validate(self):
        """Check if content is built correctly"""
        return True


class ColorOptions(Options):
    """Options related to resource display color"""
    color = properties.Color(
        'Solid color',
        default='random'
    )
    opacity = properties.Range(
        'Opacity',
        default=1.,
        min_value=0.,
        max_value=1.
    )


class MeshOptions(Options):
    """Options related to mesh display"""
    wireframe = properties.Bool(
        'Wireframe',
        default=False
    )



from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import json
import numpy as np
import os
import unittest

import properties
import steno3d


class TestResourceSurface(unittest.TestCase):

    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_surface(self):
        P = steno3d.Project()
        myVerts = np.array([[0., 0, 0], [0, 1, 0], [1, 0, 0]])
        myTriangles = np.array([[0, 1, 2]])

        S = steno3d.Surface(
            project=P,
            mesh=dict(
                vertices=myVerts,
                triangles=myTriangles,
                opts={"wireframe": True}
            ),
            opts={"opacity": 0.3, "color": "red"},
        )
        S.validate()

        # Make sure getattr/setattr working ok
        assert S.mesh is S.mesh
        assert S.mesh.vertices is S.mesh.vertices
        assert S.mesh.triangles is S.mesh.triangles

        # Make sure the shortcuts are working

        # assert S.vertices is S.mesh.vertices
        # assert S.triangles is S.mesh.triangles
        # assert S.vertices is not myVerts
        # assert S.triangles is not myTriangles
        assert np.array_equal(S.mesh.vertices, myVerts)
        assert np.array_equal(S.mesh.triangles, myTriangles)

        # Test options
        # ----Surface
        # --------Color
        assert S.opts.color == (255, 0, 0)
        S.opts.color = 'darkred'

        assert S.opts.opacity == 0.3
        S.opts.opacity = .1
        assert S.opts.opacity == .1
        S.opts = {"opacity": 0.5}
        assert S.opts.opacity == .5
        S.opts = {"color": '#FFF'}
        assert S.opts.opacity == 1

        assert S.mesh.opts.wireframe
        self.assertRaises(ValueError,
                          lambda: setattr(S.mesh.opts, 'wireframe', 'Wires'))
        self.assertRaises(ValueError,
                          lambda: setattr(S.mesh.opts, 'wireframe', 1))
        S.mesh.opts = {"wireframe": False}
        assert not S.mesh.opts.wireframe
        S.mesh.opts.wireframe = True
        assert S.mesh.opts.wireframe

        # Test triangles
        S.triangles = [[0, 0, 0], [1, 1, 1]]
        assert isinstance(S.mesh.triangles, np.ndarray)
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', -1))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', [[0, 0]]))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', [[0, 0, .5], [1, 1, 1]]))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', [[0], [0], [0]]))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', 'Three isosceles, please!'))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'triangles', [[0, 0, 1, 1], [0, 0, 0, 0]]))

        S.mesh.triangles = [[0, 0, -100], [1, 1, 1]]
        self.assertRaises(ValueError, lambda: S.validate())

        S.mesh._p_triangles = -1   # try to fake things out.
        # ensure we also call the validator
        self.assertRaises(ValueError, lambda: S.validate())

        myNewTriangles = np.array([
            [0, 1, 2], [1, 2, 3],
            [0, 1, 4], [0, 2, 4],
            [1, 3, 4], [2, 3, 4]
        ])
        S.mesh.triangles = myNewTriangles
        self.assertRaises(ValueError, lambda: S.validate())
        assert np.array_equal(S.mesh.triangles, myNewTriangles)

        # Test vertices
        myNewVerts = np.array([
            [0, 0, 0],
            [0, 0, -1],
            [0, 1, 0],
            [0, 1, 1],
            [1, .5, .5]
        ])
        S.mesh.vertices = myNewVerts
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'vertices', -1))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'vertices', [[0, 0]]))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'vertices', [[0], [0], [0]]))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'vertices', 'Just a few random points.'))
        self.assertRaises(
            ValueError,
            lambda: setattr(S.mesh, 'vertices', [[0, 0, 1, 1], [0, 0, 0, 0]]))

        # Other Constructor tests
        S = steno3d.Surface(P)
        self.assertRaises(properties.exceptions.RequiredPropertyError,
                          lambda: S.validate())
        S.mesh = S.mesh
        self.assertRaises(properties.exceptions.RequiredPropertyError,
                          lambda: S.validate())
        S.mesh.vertices = myVerts
        S.mesh.triangles = myTriangles
        S.validate()
        MOpts = dict(wireframe=True)
        M = steno3d.Mesh2D(
            vertices=myNewVerts,
            triangles=myNewTriangles,
            opts=MOpts
        )
        SOpts = dict(opacity=.5, color=[100, 100, 100])
        S = steno3d.Surface(P, mesh=M, opts=SOpts)

        assert S.mesh.vertices is not myNewVerts
        assert S.mesh.triangles is not myNewTriangles
        assert np.array_equal(S.mesh.vertices, myNewVerts)
        assert np.array_equal(S.mesh.triangles, myNewTriangles)
        assert S.opts.opacity == .5
        assert S.opts.color == (100, 100, 100)
        assert S.mesh.opts.wireframe

        # Test Data
        # This needs to be updated once data is updated
        S.mesh.triangles = myTriangles
        S.mesh.vertices = myVerts
        S.data = [{'data': [0], 'location': 'face'}]
        assert type(S.data) is list
        d0 = S.data[0]
        # copying the list and appending to it (not the same list!)
        S.data.append({'data': [2], 'location': 'vertex'})
        assert len(S.data) == 1
        # iadd is resetting the list and doing validation
        S.data += [{'data': [2], 'location': 'vertex'}]
        assert S.data[0] is d0
        from steno3d.surface import _SurfaceBinder
        assert isinstance(S.data[1], _SurfaceBinder)
        # surface.data[1] is the incorrect length
        self.assertRaises(Exception, lambda: S.validate())
        S.data[1].data = [0, 1, 2]
        assert S.data[1].data.title == ''
        assert S.data[1].data.description == ''
        assert np.all(S.data[1].data.array == [0, 1, 2])
        S.validate()

    def test_surface_mesh2dgrid(self):
        myh1 = [5., 4., 3., 2., 1., 1., 1., 1., 2., 3., 4., 5.]
        myh2 = [1., 1., 1., 1., 2., 3., 4., 5.]

        # This should error
        def f():
            P = steno3d.Project()
            S = steno3d.Surface(
                P,
                mesh=dict(
                    h1=myh1,
                    h2=myh2,
                    opts={"wireframe": True}
                ),
                opts={"opacity": 0.3, "color": "red"},
            )

        self.assertRaises(KeyError, f)

        # This should be ok
        P = steno3d.Project()
        M = steno3d.Mesh2DGrid(h1=myh1, h2=myh2, opts={"wireframe": True})
        S = steno3d.Surface(P, mesh=M, opts={"opacity": 0.3, "color": "red"})

        S.validate()

        myZ = np.random.rand(10)
        S.mesh.Z = myZ
        self.assertRaises(ValueError, lambda: S.validate())

        myZ = np.random.rand((len(myh1)+1) * (len(myh2)+1))
        S.mesh.Z = myZ
        S.validate()

        S.mesh.x0 = [[0, 0, 0], [1, 1, 1]]
        self.assertRaises(ValueError, lambda: S.validate())

    def test_teapot(self):
        basedir = os.path.realpath(os.path.dirname(steno3d.__file__))
        teapot = os.path.realpath(basedir + '/../assets/basic/teapot.json')
        with open(teapot) as f:
            data = json.loads(f.read())
        P = steno3d.Project()
        S = steno3d.Surface(
            project=P,
            mesh=dict(
                vertices=data['vertices'],
                triangles=data['triangles'],
                opts={"wireframe": True}
            ),
            opts={"opacity": 0.3, "color": "red"},
        )
        S.validate()

    def test_proj_resource_link(self):
        p0 = steno3d.Project()
        p1 = steno3d.Project()
        p2 = steno3d.Project()
        s0 = steno3d.Surface(p0)
        assert len(s0.project) == 1
        assert len(p0.resources) == 1
        s0.project = [p0, p0, p2, p1, p2]
        assert len(s0.project) == 3
        assert s0.project[0] is p0
        assert s0.project[1] is p2
        assert s0.project[2] is p1
        assert len(p0.resources) == 1
        assert len(p1.resources) == 1
        assert len(p2.resources) == 1
        s1 = steno3d.Surface(p1)
        assert len(s1.project) == 1
        assert len(p1.resources) == 2
        s2 = steno3d.Surface(p1)
        assert len(p1.resources) == 3
        s2.project += [p1]
        assert len(p1.resources) == 3
        assert len(s2.project) == 1
        p2.resources += [s0, s1, s1, s1]
        assert len(p2.resources) == 2
        assert p2.resources[0] is s0
        assert p2.resources[1] is s1
        assert len(s0.project) == 3
        assert len(s1.project) == 2


if __name__ == '__main__':
    unittest.main()
